from datetime import datetime

import pandas as pd


def date_range_inclusive(start_date: datetime, end_date: datetime, day_of_month: int = 1, freq: str = 'MS'):
    if day_of_month < 1 or day_of_month > 31:
        raise ValueError('day_of_month must be between 1 and 31')
    if end_date < start_date:
        raise ValueError('end_date must be after start_date')

    date_range = pd.date_range(
        start=start_date,
        end=end_date,
        freq=freq
    ) + pd.offsets.Day(day_of_month - 1)

    # Ensure the start_date is included
    if start_date not in date_range:
        date_range = pd.concat(
            [pd.Series([start_date]), pd.Series(date_range)]).drop_duplicates().sort_values().reset_index(
            drop=True)
    return date_range


import pytest
from datetime import datetime
import pandas as pd

from analytics.src.date_utils import date_range_inclusive


def test_valid_date_range():
    # Valid case: Start and end dates within the same year at the 1st of each month
    start_date = datetime(2023, 1, 1)
    end_date = datetime(2023, 6, 1)
    result = date_range_inclusive(start_date, end_date)
    expected = pd.concat(
        [pd.Series([start_date]), pd.Series(result)]).drop_duplicates().sort_values().reset_index(
        drop=True)
    assert all(expected == result)


def test_day_of_month_adjustment():
    # Ensure the `day_of_month` parameter is correctly applied
    start_date = datetime(2023, 1, 15)
    end_date = datetime(2023, 6, 15)
    day_of_month = 15

    result = date_range_inclusive(start_date, end_date, day_of_month=day_of_month)
    expected = pd.concat(
        [pd.Series([start_date]), pd.Series(result)]).drop_duplicates().sort_values().reset_index(
        drop=True)
    assert all(expected == result)


def test_include_start_date():
    # Test that the start_date is always included
    start_date = datetime(2023, 1, 10)
    end_date = datetime(2023, 6, 10)
    day_of_month = 10
    result = date_range_inclusive(start_date, end_date, day_of_month)
    assert start_date in [pd.to_datetime(r) for r in result]


def test_invalid_day_of_month_below_range():
    # Check that ValueError is raised for day_of_month < 1
    start_date = datetime(2023, 1, 1)
    end_date = datetime(2023, 6, 1)
    with pytest.raises(ValueError, match='day_of_month must be between 1 and 31'):
        date_range_inclusive(start_date, end_date, day_of_month=0)


def test_invalid_day_of_month_above_range():
    # Check that ValueError is raised for day_of_month > 31
    start_date = datetime(2023, 1, 1)
    end_date = datetime(2023, 6, 1)
    with pytest.raises(ValueError, match='day_of_month must be between 1 and 31'):
        date_range_inclusive(start_date, end_date, day_of_month=32)


def test_end_date_before_start_date():
    # Check that ValueError is raised if end_date < start_date
    start_date = datetime(2023, 6, 1)
    end_date = datetime(2023, 1, 1)
    with pytest.raises(ValueError, match='end_date must be after start_date'):
        date_range_inclusive(start_date, end_date)


def test_single_date_range():
    # Test the edge case where start_date == end_date
    start_date = datetime(2023, 1, 1)
    end_date = datetime(2023, 1, 1)
    expected = pd.date_range(start=start_date, end=end_date, freq='MS') + pd.offsets.Day(0)
    result = date_range_inclusive(start_date, end_date)
    pd.testing.assert_index_equal(result, expected)



from datetime import datetime
from typing import List

import pandas as pd
from pandas import DatetimeIndex

from analytics.src.counterparty_info import CounterPartyInfo
from analytics.src.date_utils import date_range_inclusive


class CashFlow:
    def __init__(self,
                 name: str,
                 counterparties: CounterPartyInfo,
                 start_date: datetime | str,
                 end_date: datetime | str = datetime(2200, 12, 31),
                 frequency: str = 'MS',
                 day_of_month: int = None,
                 amount: float = 0.0):
        if isinstance(start_date, str):
            start_date = datetime.strptime(start_date, '%Y-%m-%d')
        if isinstance(end_date, str):
            end_date = datetime.strptime(end_date, '%Y-%m-%d')
        if day_of_month is None:
            day_of_month = start_date.date().day
        if day_of_month < 1 or day_of_month > 31:
            raise ValueError('day_of_month must be between 1 and 31')
        if end_date < start_date:
            raise ValueError('end_date must be after start_date')
        self.__counterparties__ = counterparties
        self.__start_date__ = start_date
        self.__end_date__ = end_date
        self.__frequency__ = frequency
        self.__day_of_month__ = day_of_month
        self.__amount__ = amount
        self.__cashflow_name__ = name

        self.__settled__: pd.DataFrame = pd.DataFrame(columns=['Date', 'Amount'])
        self.__overridden__: pd.DataFrame = pd.DataFrame(columns=['Date', 'Amount'])
        self.__cancelled__: List[datetime] = []

    @property
    def name(self) -> str:
        return self.__cashflow_name__

    @property
    def counterparties(self) -> CounterPartyInfo:
        return self.__counterparties__

    @property
    def start_date(self) -> datetime:
        return self.__start_date__

    @property
    def end_date(self) -> datetime:
        return self.__end_date__

    @property
    def frequency(self) -> str:
        return self.__frequency__

    @property
    def amount(self) -> float:
        return self.__amount__

    @property
    def day_of_month(self) -> int:
        return self.__day_of_month__

    @property
    def cash_flow_dates(self) -> DatetimeIndex:
        return date_range_inclusive(
            start_date=self.__start_date__,
            end_date=self.__end_date__,
            day_of_month=self.__day_of_month__,
            freq=self.__frequency__)

    def add_settlement(self, date: datetime, amount: float):
        if date not in self.__settled__['Date'].values:
            self.__settled__ = pd.concat([self.__settled__, pd.DataFrame({'Date': date, 'Amount': amount}, index=[0])])
        else:
            self.__settled__.loc[self.__settled__['Date'] == date, 'Amount'] += amount

    def add_overridden(self, date: datetime, amount: float):
        if date not in self.__overridden__['Date'].values:
            self.__overridden__ = pd.concat(
                [self.__overridden__, pd.DataFrame({'Date': date, 'Amount': amount}, index=[0])])
        else:
            self.__overridden__.loc[self.__overridden__['Date'] == date, 'Amount'] += amount

    def add_cancelled(self, date: datetime):
        if date not in self.__cancelled__:
            self.__cancelled__.append(date)

    def cashflows(self, from_date: datetime | str = None) -> pd.DataFrame:

        if isinstance(from_date, str):
            from_date = datetime.strptime(from_date, '%Y-%m-%d')

        df = pd.DataFrame(self.cash_flow_dates, columns=['Date'])
        df['Amount'] = self.amount
        df['Status'] = None

        # Merge in Overrides
        df = df.merge(
            self.__overridden__,
            on="Date",
            how="left",
            suffixes=("", "_overridden"),
        )
        df["Amount"] = df["Amount_overridden"].combine_first(df["Amount"])
        df.drop(columns=["Amount_overridden"], inplace=True)

        # Merge in Settlements

        df = df.merge(
            self.__settled__,
            on="Date",
            how="outer",
            suffixes=("", "_settled"),
        )
        df["Amount"] = df["Amount_settled"].combine_first(df["Amount"])
        df.drop(columns=["Amount_settled"], inplace=True)

        df['Source'] = self.counterparties.source
        df['Target'] = self.counterparties.target
        df['Cashflow'] = self.name
        df.loc[df['Date'].isin(self.__overridden__['Date']), 'Status'] = 'Overridden'
        df.loc[df['Date'].isin(self.__settled__['Date']), 'Status'] = 'Settled'

        # Set Cancellations
        df.loc[df['Date'].isin(self.__cancelled__), 'Status'] = 'Cancelled'
        df.loc[df['Date'].isin(self.__cancelled__), 'Amount'] = 0

        if from_date is not None:
            df = df[df['Date'] >= from_date]

        return df

    def to_dict(self):
        return {
            "cashflow_name": self.__cashflow_name__,
            "counterparties": self.__counterparties__.to_dict(),
            "start_date": self.__start_date__.strftime('%Y-%m-%d'),
            "end_date": self.__end_date__.strftime('%Y-%m-%d'),
            "frequency": self.__frequency__,
            "amount": self.__amount__,
            "day_of_month": self.__day_of_month__,
            "settled": self.__settled__.map(
                lambda x: x.strftime('%Y-%m-%d') if isinstance(x, pd.Timestamp) else x
            )
            .to_dict(orient='records'),
            "overridden": self.__overridden__.map(
                lambda x: x.strftime('%Y-%m-%d') if isinstance(x, pd.Timestamp) else x
            )
            .to_dict(orient='records'),
            "cancelled": [c.strftime('%Y-%m-%d') for c in self.__cancelled__]

        }

    @staticmethod
    def from_dict(data):
        """
        Create a CashFlow object from a dictionary representation.
        """
        # Parse start_date and end_date as datetime objects
        start_date = datetime.strptime(data["start_date"], '%Y-%m-%d') if "start_date" in data else None
        end_date = datetime.strptime(data["end_date"], '%Y-%m-%d') if "end_date" in data else None

        # Parse counterparties (assumes it's stored as a dictionary and the corresponding class has a from_dict method)
        counterparties = CounterPartyInfo.from_dict(data["counterparties"])

        # Parse settled and overridden (convert lists of records back into DataFrames)
        settled = pd.DataFrame(data["settled"]) if "settled" in data else None
        overridden = pd.DataFrame(data["overridden"]) if "overridden" in data else None

        # Parse cancelled dates back into datetime objects
        cancelled = [datetime.strptime(c, '%Y-%m-%d') for c in data["cancelled"]] if "cancelled" in data else []

        cflw = CashFlow(
            name=data["cashflow_name"],
            counterparties=counterparties,
            start_date=start_date,
            end_date=end_date,
            frequency=data["frequency"],
            amount=data["amount"],
            day_of_month=data["day_of_month"],
        )

        for idx, row in settled.iterrows():
            cflw.add_settlement(pd.to_datetime(row['Date']), row['Amount'])

        for idx, row in overridden.iterrows():
            cflw.add_overridden(pd.to_datetime(row['Date']), row['Amount'])

        for c in cancelled:
            cflw.add_cancelled(c)
        return cflw



import json


class CounterPartyInfo:
    def __init__(self, source: str, target: str):
        self.source = source
        self.target = target

    def to_dict(self):
        return {
            "source": self.source,
            "target": self.target,
        }

    @staticmethod
    def from_dict(data: dict):
        """
        Deserialize a dictionary to create a CounterPartyInfo instance.
        """
        return CounterPartyInfo(
            source=data.get("source"),  # Retrieve the 'source' key from the dictionary
            target=data.get("target"),  # Retrieve the 'target' key from the dictionary
        )



import pytest
import pandas as pd
from datetime import datetime

from analytics.src.cashflow import CashFlow
from analytics.src.counterparty_info import CounterPartyInfo
from analytics.src.date_utils import date_range_inclusive


@pytest.fixture
def sample_cashflow():
    # A sample `CashFlow` object to use across tests
    return CashFlow(
        name="Test CashFlow",
        counterparties=CounterPartyInfo("source","target"),
        start_date=datetime(2023, 1, 1),
        end_date=datetime(2023, 6, 30),
        amount=1000
    )



class TestCashFlow:
    def setup_method(self):
        self.counterparties = CounterPartyInfo("source","target")

    def test_default_arguments(self):
        """Test to verify default arguments are handled correctly."""
        start_date = datetime(2023, 1, 1)
        cash_flow = CashFlow("test",self.counterparties, start_date=start_date)

        assert cash_flow.start_date == start_date
        assert cash_flow.end_date == datetime(2200, 12, 31)
        assert cash_flow.frequency == 'MS'

        # Verify that correct cash flow dates are generated.
        expected_dates = pd.date_range(start=start_date, end=datetime(2200, 12, 31), freq='MS') \
                         + pd.offsets.Day(0)
        assert all(cash_flow.cash_flow_dates == expected_dates)

    def test_custom_end_date(self):
        """Test when a custom end date is provided."""
        start_date = datetime(2023, 1, 1)
        end_date = datetime(2023, 6, 1)
        cash_flow = CashFlow("test",self.counterparties, start_date=start_date, end_date=end_date)

        assert cash_flow.start_date == start_date
        assert cash_flow.end_date == end_date

        expected_dates = date_range_inclusive(start_date=cash_flow.start_date,
                                              end_date=cash_flow.end_date,
                                              day_of_month=1,
                                              freq=cash_flow.frequency)
        assert all(cash_flow.cash_flow_dates == expected_dates)

    def test_default_arguments_as_str(self):
        """Test to verify default arguments are handled correctly."""
        start_date = '2023-01-01'
        end_date = '2200-12-31'

        cash_flow = CashFlow("test",self.counterparties, start_date=start_date, end_date=end_date)

        assert cash_flow.start_date == datetime(2023, 1, 1)
        assert cash_flow.end_date == datetime(2200, 12, 31)
        assert cash_flow.frequency == 'MS'

        # Verify that correct cash flow dates are generated.
        expected_dates = date_range_inclusive(start_date=cash_flow.start_date,
                                              end_date=cash_flow.end_date,day_of_month=1,
                                              freq=cash_flow.frequency)
        assert all(cash_flow.cash_flow_dates == expected_dates)

    def test_custom_day_of_month(self):
        """Test the handling of custom day_of_month."""
        start_date = datetime(2023, 1, 1)
        end_date = datetime(2023, 4, 1)
        day_of_month = 15
        cash_flow = CashFlow("test",self.counterparties, start_date=start_date, end_date=end_date, day_of_month=day_of_month)

        assert cash_flow.start_date == start_date
        assert cash_flow.end_date == end_date

        # Verify that cash flow generates dates with the correct day_of_month.
        expected_dates = date_range_inclusive(start_date, end_date, day_of_month=day_of_month)
        assert all(cash_flow.cash_flow_dates == expected_dates)

    def test_invalid_day_of_month(self):
        """Test for ValueError when day_of_month is invalid (e.g., <1 or >31)."""
        start_date = datetime(2023, 1, 1)

        with pytest.raises(ValueError):
            CashFlow("test",self.counterparties, start_date=start_date, day_of_month=0)  # Invalid day_of_month

        with pytest.raises(ValueError):
            CashFlow("test",self.counterparties, start_date=start_date, day_of_month=32)  # Invalid day_of_month

    def test_invalid_end_date_less_than_start_date(self):
        """Test for ValueError when day_of_month is invalid (e.g., <1 or >31)."""
        start_date = datetime(2023, 1, 1)
        end_date = datetime(2022, 12, 31)

        with pytest.raises(ValueError):
            CashFlow("test",self.counterparties, start_date=start_date, end_date=end_date)  # Invalid day_of_month

    def test_large_date_range(self):
        """Test the generation of cash flow dates over a very large range."""
        start_date = datetime(2023, 1, 1)
        end_date = datetime(2025, 12, 31)
        cash_flow = CashFlow("test",self.counterparties, start_date=start_date, end_date=end_date)

        assert cash_flow.start_date == start_date
        assert cash_flow.end_date == end_date

        expected_dates = pd.date_range(start=start_date, end=end_date, freq='MS') + pd.offsets.Day(0)
        assert all(cash_flow.cash_flow_dates == expected_dates)
        assert len(cash_flow.cash_flow_dates) == len(expected_dates)

    def test_edge_case_end_of_month(self):
        """Test for cash flows generating dates at the end of each month."""
        start_date = datetime(2023, 2, 28)
        end_date = datetime(2023, 6, 28)
        day_of_month = 28
        cash_flow = CashFlow("test",self.counterparties, start_date=start_date, end_date=end_date, day_of_month=day_of_month)

        assert cash_flow.start_date == start_date
        assert cash_flow.end_date == end_date

        # Since setting day_of_month=28 in February might go beyond the range, verify edge handling
        expected_dates = date_range_inclusive(cash_flow.start_date, cash_flow.end_date, day_of_month=cash_flow.day_of_month)
        assert all(cash_flow.cash_flow_dates == expected_dates)

    def test_amount(self):
        start_date = datetime(2023, 1, 1)
        end_date = datetime(2023, 6, 1)
        cash_flow = CashFlow("test",self.counterparties, start_date=start_date, end_date=end_date,amount=1000)

        assert cash_flow.amount == 1000

    def test_name(self):
        start_date = datetime(2023, 1, 1)
        end_date = datetime(2023, 6, 1)
        cash_flow = CashFlow("test",self.counterparties, start_date=start_date, end_date=end_date,amount=1000)

        assert cash_flow.name == 'test'

    def test_day_of_month(self):
        start_date = datetime(2023, 1, 1)
        end_date = datetime(2023, 6, 1)
        cash_flow = CashFlow("test",self.counterparties, start_date=start_date, end_date=end_date,amount=1000, day_of_month=15)

        assert cash_flow.day_of_month == 15

    def test_counterparties(self):
        start_date = datetime(2023, 1, 1)
        end_date = datetime(2023, 6, 1)
        cash_flow = CashFlow("test",self.counterparties, start_date=start_date, end_date=end_date,amount=1000)

        assert cash_flow.counterparties.source == 'source'
        assert cash_flow.counterparties.target == 'target'

    def test_cashflows_valid_dates(self):
        """
        Test the cashflows method generates the correct cash flow schedule for valid input.
        """
        start_date = datetime(2023, 1, 15)
        end_date = datetime(2023, 3, 15)
        cashflow = CashFlow(
            "Test",
            self.counterparties,
            start_date=start_date,
            end_date=end_date,
            frequency="MS",
            amount=100,
            day_of_month=15
        )
        result = cashflow.cashflows()
        assert len(result) == 3, "Expected 3 cash flow entries for a monthly frequency"
        assert pd.to_datetime(result["Date"].values[0]) == datetime(2023, 1, 15), "First entry date should match 15th Jan 2023"
        assert pd.to_datetime(result["Date"].values[-1]) == datetime(2023, 3, 15), "Last entry date should match 15th Mar 2023"
        assert all([entry == 100 for entry in result['Amount']]), "Every entry should have the correct amount"

    def test_cashflows_edge_case_end_of_month(self):
        """
        Test the cashflows method when the start date is at the end of the month.
        """
        start_date = datetime(2023, 1, 15)
        end_date = datetime(2023, 3, 15)
        cashflow = CashFlow(
            "End of Month Test",
            self.counterparties,
            start_date=start_date,
            end_date=end_date,
            frequency="MS",
            amount=200,
            day_of_month=15
        )
        result = cashflow.cashflows()
        assert len(result) == 3, "Expected 3 cash flow entries for a monthly frequency"
        assert all(pd.to_datetime(entry).day == 15 for entry in result['Date']), "All cash flow dates should fall on the 31st"





import unittest

from analytics.src.counterparty_info import CounterPartyInfo


class TestCounterPartyInfo(unittest.TestCase):
    def test_initialization(self):
        """Test that the source and target are properly initialized."""
        source = "Bank A"
        target = "Bank B"

        counterparty_info = CounterPartyInfo(source, target)

        self.assertEqual(counterparty_info.source, source)
        self.assertEqual(counterparty_info.target, target)

    def test_empty_initialization(self):
        """Test initialization with empty strings."""
        source = ""
        target = ""

        counterparty_info = CounterPartyInfo(source, target)

        self.assertEqual(counterparty_info.source, source)
        self.assertEqual(counterparty_info.target, target)

    def test_none_initialization(self):
        """Test initialization with None as parameters."""
        source = None
        target = None

        counterparty_info = CounterPartyInfo(source, target)

        self.assertIsNone(counterparty_info.source)
        self.assertIsNone(counterparty_info.target)

    def test_mutability(self):
        """Test that the attributes can be changed after initialization."""
        counterparty_info = CounterPartyInfo(source="Bank A", target="Bank B")

        # Change values
        counterparty_info.source = "Bank X"
        counterparty_info.target = "Bank Y"

        self.assertEqual(counterparty_info.source, "Bank X")
        self.assertEqual(counterparty_info.target, "Bank Y")
